import pandas as pd
import numpy as np
import umap
import hdbscan
import math
import optuna
import logging
from sklearn.metrics import silhouette_score


class ClusteringEngine:
    def __init__(self, 
        dims=3, 
        min_clusters=3, 
        max_clusters=25, 
        n_trials=50, 
        random_state=None,
        embedding_col_name="embedding_vector",
        
        umap_n_neighbors_min=2,
        umap_n_neighbors_max=25,
        umap_min_dist_min=0.0,
        umap_min_dist_max=0.1,
        umap_spread_min=1.0,
        umap_spread_max=10.0,
        umap_learning_rate_min=0.08,
        umap_learning_rate_max=1.0,
        umap_min_dims=2,
        umap_max_dims=20,
        umap_metric="cosine",
        
        hdbscan_min_cluster_size_multiplier_min=0.005,
        hdbscan_min_cluster_size_multiplier_max=0.025,
        hdbscan_min_samples_min=2,
        hdbscan_min_samples_max=50,
        hdbscan_epsilon_min=0.0,
        hdbscan_epsilon_max=1.0,
        hdbscan_metric="euclidean",
        hdbscan_cluster_selection_method="eom",
        
        optuna_jobs=-1,
        hdbscan_outlier_threshold=90
    ):
        """
        Initialize the optimizer with UMAP and HDBSCAN hyperparameter settings.

        Args:
            - dims (int or None): Number of dimensions for UMAP reduction. If None, a wider search is performed.
            - min_clusters (int): Minimum acceptable number of clusters.
            - max_clusters (int): Maximum acceptable number of clusters.
            - n_trials (int): Number of optimization trials for hyperparameter tuning.
            - random_state: Seed for random number generators to ensure reproducibility.
            - embedding_col_name (str): Name of the column containing embedding vectors.
            - umap_n_neighbors_min (int): Minimum value for UMAP's n_neighbors parameter.
            - umap_n_neighbors_max (int): Maximum value for UMAP's n_neighbors parameter.
            - umap_min_dist_min (float): Minimum value for UMAP's min_dist parameter.
            - umap_min_dist_max (float): Maximum value for UMAP's min_dist parameter.
            - umap_spread_min (float): Minimum value for UMAP's spread parameter.
            - umap_spread_max (float): Maximum value for UMAP's spread parameter.
            - umap_learning_rate_min (float): Minimum value for UMAP's learning_rate parameter.
            - umap_learning_rate_max (float): Maximum value for UMAP's learning_rate parameter.
            - umap_min_dims (int): Minimum number of dimensions for UMAP reduction.
            - umap_max_dims (int): Maximum number of dimensions for UMAP reduction.
            - umap_metric (str): Distance metric to use for UMAP reduction.
            - hdbscan_min_cluster_size_multiplier_min (float): Minimum multiplier for HDBSCAN's min_cluster_size.
            - hdbscan_min_cluster_size_multiplier_max (float): Maximum multiplier for HDBSCAN's min_cluster_size.
            - hdbscan_min_samples_min (int): Minimum value for HDBSCAN's min_samples parameter.
            - hdbscan_min_samples_max (int): Maximum value for HDBSCAN's min_samples parameter.
            - hdbscan_epsilon_min (float): Minimum value for HDBSCAN's cluster_selection_epsilon parameter.
            - hdbscan_epsilon_max (float): Maximum value for HDBSCAN's cluster_selection_epsilon parameter.
            - hdbscan_metric (str): Distance metric to use for HDBSCAN clustering.
            - hdbscan_cluster_selection_method (str): Method to use for selecting clusters in HDBSCAN.
            - optuna_jobs (int): Number of parallel jobs to run during Optuna optimization. Set to -1 for all CPUs.
            - hdbscan_outlier_threshold (int): Percentile threshold for HDBSCAN outlier detection.
            - log_level: Logging level (e.g., logging.INFO, logging.DEBUG)
        """
        # -------------------------------
        # Assign initialization parameters to instance attributes.
        # -------------------------------
        self.dims: int = dims
        self.min_clusters: int = min_clusters
        self.max_clusters: int = max_clusters
        self.n_trials: int = n_trials
        self.random_state: int|None = random_state
        self.embedding_col_name: str = embedding_col_name
        
        self.umap_n_neighbors_min: int = umap_n_neighbors_min
        self.umap_n_neighbors_max: int = umap_n_neighbors_max
        self.umap_min_dist_min: float = umap_min_dist_min
        self.umap_min_dist_max: float = umap_min_dist_max
        self.umap_spread_min: float = umap_spread_min
        self.umap_spread_max: float = umap_spread_max
        self.umap_learning_rate_min: float = umap_learning_rate_min
        self.umap_learning_rate_max: float = umap_learning_rate_max
        self.umap_min_dims: int = umap_min_dims
        self.umap_max_dims: int = umap_max_dims
        self.umap_metric: str = umap_metric
        
        self.hdbscan_min_cluster_size_multiplier_min: float = hdbscan_min_cluster_size_multiplier_min
        self.hdbscan_min_cluster_size_multiplier_max: float = hdbscan_min_cluster_size_multiplier_max
        self.hdbscan_min_samples_min: int = hdbscan_min_samples_min
        self.hdbscan_min_samples_max: int = hdbscan_min_samples_max
        self.hdbscan_epsilon_min: float = hdbscan_epsilon_min
        self.hdbscan_epsilon_max: float = hdbscan_epsilon_max
        self.hdbscan_metric: str = hdbscan_metric
        self.hdbscan_cluster_selection_method: str = hdbscan_cluster_selection_method
        
        self.optuna_jobs: int = optuna_jobs
        self.hdbscan_outlier_threshold: int = hdbscan_outlier_threshold

        # -------------------------------
        # Logger Initialization: Create and configure a logger for the ClusteringEngine instance.
        # -------------------------------
        self.logger = logging.getLogger(self.__class__.__name__)
        
        self.logger.debug(
            "ClusteringEngine initialized with dims=%s, min_clusters=%s, max_clusters=%s, n_trials=%s, random_state=%s",
            self.dims, 
            self.min_clusters, 
            self.max_clusters, 
            self.n_trials, 
            self.random_state
        )

    def _create_models(self, trial, num_data_pts):
        """
        Create UMAP and HDBSCAN models with hyperparameters suggested by the Optuna trial.

        The UMAP parameters are chosen based on trial suggestions. The search range for UMAP's n_components
        is restricted to [2, 3] if self.dims is not None, otherwise, it spans a wider range.
        HDBSCAN parameters are also suggested based on the number of data points.

        Args:
            trial (optuna.trial.Trial): Optuna trial object to sample hyperparameters.
            num_data_pts (int): Number of data points in the dataset.

        Returns:
            tuple: A tuple containing:
                - umap.UMAP: UMAP model instance with suggested parameters.
                - hdbscan.HDBSCAN: HDBSCAN model instance with suggested parameters.
                - dict: Dictionary of UMAP parameters.
                - dict: Dictionary of HDBSCAN parameters.
        """
        self.logger.debug("Creating models for trial with num_data_pts=%d", num_data_pts)
        # -------------------------------
        # UMAP Parameter Selection:
        # Determine the number of components and suggest hyperparameters using the trial.
        # -------------------------------
        if self.dims is None:
            umap_n_components = trial.suggest_int("umap_n_components", self.umap_min_dims, self.umap_max_dims)
        else:
            umap_n_components = self.dims
            
        umap_params = {
            "n_neighbors": trial.suggest_int("umap_n_neighbors", self.umap_n_neighbors_min, self.umap_n_neighbors_max),
            "min_dist": trial.suggest_float("umap_min_dist", self.umap_min_dist_min, self.umap_min_dist_max),
            "spread": trial.suggest_float("umap_spread", self.umap_spread_min, self.umap_spread_max),
            "metric": self.umap_metric,
            "random_state": self.random_state,
            "learning_rate": trial.suggest_float("umap_learning_rate", self.umap_learning_rate_min, self.umap_learning_rate_max),
            "init": "spectral",
            "n_components": umap_n_components,
        }
        self.logger.debug("UMAP parameters: %s", umap_params)
        
        # -------------------------------
        # HDBSCAN Parameter Selection:
        # Suggest hyperparameters for HDBSCAN based on the number of data points.
        # -------------------------------
        hdbscan_params = {
            "min_cluster_size": trial.suggest_int(
                "hdbscan_min_cluster_size",
                math.ceil(self.hdbscan_min_cluster_size_multiplier_min * num_data_pts),
                math.ceil(self.hdbscan_min_cluster_size_multiplier_max * num_data_pts)
            ),
            "min_samples": trial.suggest_int("hdbscan_min_samples", self.hdbscan_min_samples_min, self.hdbscan_min_samples_max),
            "cluster_selection_epsilon": trial.suggest_float("hdbscan_epsilon", self.hdbscan_epsilon_min, self.hdbscan_epsilon_max),
            "metric": self.hdbscan_metric,
            "cluster_selection_method": self.hdbscan_cluster_selection_method,
        }
        self.logger.debug("HDBSCAN parameters: %s", hdbscan_params)
        
        # -------------------------------
        # Return the constructed UMAP and HDBSCAN models along with their parameter dictionaries.
        # -------------------------------
        return (
            umap.UMAP(**umap_params),
            hdbscan.HDBSCAN(**hdbscan_params),
            umap_params,
            hdbscan_params,
        )
    
    def _default_models(self, num_data_pts):
        """
        Create UMAP and HDBSCAN models using a set of predefined default hyperparameters.
        This fallback is used when optimization fails to find any valid Pareto-optimal trials.

        Args:
            num_data_pts (int): Number of data points in the dataset.

        Returns:
            tuple: A tuple containing:
                - umap.UMAP: UMAP model instance with default parameters.
                - hdbscan.HDBSCAN: HDBSCAN model instance with default parameters.
                - dict: Dictionary of default UMAP parameters.
                - dict: Dictionary of default HDBSCAN parameters.
        """
        self.logger.debug("Creating default models using predefined hyperparameters for fallback.")
        # -------------------------------
        # Determine default n_components for UMAP.
        # -------------------------------
        umap_n_components = self.dims if self.dims is not None else 3
        
        # -------------------------------
        # Compute default UMAP parameters as the average of the min and max values.
        # -------------------------------
        umap_params = {
            "n_neighbors": (self.umap_n_neighbors_min + self.umap_n_neighbors_max) // 2,
            "min_dist": (self.umap_min_dist_min + self.umap_min_dist_max) / 2,
            "spread": (self.umap_spread_min + self.umap_spread_max) / 2,
            "metric": self.umap_metric,
            "random_state": self.random_state,
            "learning_rate": (self.umap_learning_rate_min + self.umap_learning_rate_max) / 2,
            "init": "spectral",
            "n_components": umap_n_components,
        }
        # -------------------------------
        # Compute default HDBSCAN parameters similarly.
        # -------------------------------
        hdbscan_params = {
            "min_cluster_size": math.ceil(((self.hdbscan_min_cluster_size_multiplier_min + self.hdbscan_min_cluster_size_multiplier_max) / 2) * num_data_pts),
            "min_samples": (self.hdbscan_min_samples_min + self.hdbscan_min_samples_max) // 2,
            "cluster_selection_epsilon": (self.hdbscan_epsilon_min + self.hdbscan_epsilon_max) / 2,
            "metric": self.hdbscan_metric,
            "cluster_selection_method": self.hdbscan_cluster_selection_method,
        }
        self.logger.debug("Default UMAP parameters: %s", umap_params)
        self.logger.debug("Default HDBSCAN parameters: %s", hdbscan_params)
        
        # -------------------------------
        # Instantiate the UMAP and HDBSCAN models with the default parameters.
        # -------------------------------
        umap_model = umap.UMAP(**umap_params)
        hdbscan_model = hdbscan.HDBSCAN(**hdbscan_params)
        return umap_model, hdbscan_model, umap_params, hdbscan_params

    def _compute_metrics(self, reduced_data, labels):
        """
        Compute clustering metrics including silhouette score and negative noise ratio.

        This function calculates the silhouette score for non-noise points (where noise is denoted by label -1)
        and computes the negative noise ratio. Returns None if there's only one cluster or too few non-noise points.

        Args:
            reduced_data (np.ndarray): Data after dimensionality reduction.
            labels (np.ndarray): Cluster labels assigned by HDBSCAN, with -1 indicating noise.

        Returns:
            dict or None: Dictionary containing 'silhouette' and 'neg_noise' if valid, otherwise None.
        """
        # -------------------------------
        # Identify non-noise points (labels != -1).
        # -------------------------------
        mask = labels != -1
        self.logger.debug("Computed non-noise mask: %s (sum=%d)", mask, np.sum(mask))
        
        # -------------------------------
        # Check if there are enough clusters and non-noise points to compute metrics.
        # -------------------------------
        if len(np.unique(labels)) <= 1 or np.sum(mask) < 2:
            self.logger.debug("Not enough clusters or non-noise points to compute metrics.")
            return None
        
        # -------------------------------
        # Compute the silhouette score for the non-noise data.
        # -------------------------------
        silhouette = silhouette_score(reduced_data[mask], labels[mask], metric="euclidean")
        
        # -------------------------------
        # Calculate the negative noise ratio (fewer noise points is better).
        # -------------------------------
        neg_noise = -((labels == -1).sum() / len(labels))
        self.logger.debug("Computed metrics: silhouette=%.4f, neg_noise=%.4f", silhouette, neg_noise)
        return {"silhouette": silhouette, "neg_noise": neg_noise}

    def _triple_objective(self, trial, embeddings):
        """
        Objective function for Optuna optimization combining three metrics.

        This function applies UMAP and HDBSCAN with hyperparameters from the trial, computes clustering metrics,
        and returns a triple of objectives:
            - silhouette score (to maximize)
            - negative noise ratio (to maximize)
            - negative number of clusters (to minimize the number of clusters)

        If any errors occur or constraints are not met (e.g., cluster count outside allowed range), it returns
        a triple of -infinity values to indicate an invalid trial.

        Args:
            trial (optuna.trial.Trial): Optuna trial for hyperparameter suggestions.
            embeddings (np.ndarray): Array of embedding vectors.

        Returns:
            list: A list containing [silhouette, negative noise ratio, negative number of clusters].
        """
        self.logger.debug("Starting triple-objective evaluation for trial number: %s", trial.number if hasattr(trial, 'number') else 'N/A')
        try:
            # -------------------------------
            # Create models using trial-specific hyperparameters.
            # -------------------------------
            umap_model, hdbscan_model, _, _ = self._create_models(trial, len(embeddings))
            self.logger.debug("Models created for trial %s", trial.number if hasattr(trial, 'number') else 'N/A')
            
            # -------------------------------
            # Apply UMAP dimensionality reduction.
            # -------------------------------
            reduced_data = umap_model.fit_transform(embeddings)
            self.logger.debug("UMAP reduction completed for trial %s", trial.number if hasattr(trial, 'number') else 'N/A')
            
            # -------------------------------
            # Predict cluster labels using HDBSCAN.
            # -------------------------------
            labels = hdbscan_model.fit_predict(reduced_data) # type: ignore
            self.logger.debug("HDBSCAN clustering produced labels: %s", labels)
            
            # -------------------------------
            # Compute clustering metrics (silhouette and negative noise ratio).
            # -------------------------------
            metrics_result = self._compute_metrics(reduced_data, labels)
            if metrics_result is None:
                self.logger.debug("Metrics result invalid for trial %s; returning -inf objectives.", trial.number if hasattr(trial, 'number') else 'N/A')
                return [float("-inf")] * 3

            # -------------------------------
            # Extract metrics and compute number of clusters (excluding noise).
            # -------------------------------
            s = metrics_result["silhouette"]
            neg_noise = metrics_result["neg_noise"]
            k = len(set(labels) - {-1})
            self.logger.debug("Trial %s metrics: silhouette=%.4f, neg_noise=%.4f, clusters=%d", trial.number if hasattr(trial, 'number') else 'N/A', s, neg_noise, k)
            
            # -------------------------------
            # Enforce cluster count constraints.
            # -------------------------------
            if k < self.min_clusters or k > self.max_clusters:
                self.logger.debug("Trial %s failed cluster count constraints (k=%d).", trial.number if hasattr(trial, 'number') else 'N/A', k)
                return [float("-inf")] * 3
            
            neg_k = -k  # Fewer clusters is better.
            return [s, neg_noise, neg_k]
        except Exception as e:
            # -------------------------------
            # Log any error that occurs and mark this trial as invalid.
            # -------------------------------
            self.logger.error("Trial failed with error: %s", str(e))
            return [float("-inf")] * 3

    def _euclidean_distance_3d(self, x1, y1, z1, x2, y2, z2):
        """
        Compute the Euclidean distance between two points in 3D space.

        Args:
            x1, y1, z1 (float): Coordinates of the first point.
            x2, y2, z2 (float): Coordinates of the second point.

        Returns:
            float: The Euclidean distance between the two points.
        """
        # -------------------------------
        # Calculate the Euclidean distance using the standard 3D distance formula.
        # -------------------------------
        distance = math.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2 + (z1 - z2) ** 2)
        self.logger.debug("Computed Euclidean distance between (%.4f, %.4f, %.4f) and (%.4f, %.4f, %.4f): %.4f",
                            x1, y1, z1, x2, y2, z2, distance)
        return distance

    def _get_best_solution(self, study, pareto_trials):
        """
        Select the best trial from the Pareto optimal solutions using the TOPSIS method.

        The Technique for Order Preference by Similarity to Ideal Solution (TOPSIS) is applied on the Pareto front.
        Each trial's metrics (silhouette, negative noise, and negative number of clusters) are normalized, and
        distances to the ideal and anti-ideal solutions are computed. The trial with the highest TOPSIS score is selected.

        Args:
            study (optuna.study.Study): The Optuna study containing all trials.
            pareto_trials (list): List of Pareto-optimal trials with valid metric values.

        Returns:
            tuple: A tuple containing:
                - best_trial (optuna.trial.FrozenTrial): The selected best trial.
                - str: A string indicating the selection method ("pareto_topsis").

        Raises:
            ValueError: If no valid Pareto optimal solutions are found.
        """
        self.logger.debug("Selecting best solution from %d Pareto trials.", len(pareto_trials))
        if pareto_trials:
            try:
                # -------------------------------
                # Extract and collect metrics from each Pareto trial.
                # -------------------------------
                trial_details = []
                for t in pareto_trials:
                    trial_details.append({
                        "trial": t,
                        "silhouette": t.values[0],
                        "neg_noise": t.values[1],
                        "neg_k": t.values[2],
                    })
                self.logger.debug("Extracted trial details: %s", trial_details)
                
                # -------------------------------
                # Collect lists of each individual metric for normalization.
                # -------------------------------
                sil_vals = [d["silhouette"] for d in trial_details]
                noise_vals = [d["neg_noise"] for d in trial_details]
                k_vals = [d["neg_k"] for d in trial_details]

                # -------------------------------
                # Define a normalization function (Euclidean norm factor).
                # -------------------------------
                def norm_factor(vals):
                    return math.sqrt(sum(v * v for v in vals))

                # -------------------------------
                # Calculate normalization factors for each metric.
                # -------------------------------
                sil_norm = norm_factor(sil_vals)
                noise_norm = norm_factor(noise_vals)
                k_norm = norm_factor(k_vals)
                self.logger.debug("Normalization factors: sil_norm=%.4f, noise_norm=%.4f, k_norm=%.4f", sil_norm, noise_norm, k_norm)

                # -------------------------------
                # Normalize each trial's metrics.
                # -------------------------------
                normalized = []
                for d in trial_details:
                    s_norm = d["silhouette"] / sil_norm if sil_norm != 0 else 0
                    n_norm = d["neg_noise"] / noise_norm if noise_norm != 0 else 0
                    k_norm_val = d["neg_k"] / k_norm if k_norm != 0 else 0
                    normalized.append({**d, "s_norm": s_norm, "n_norm": n_norm, "k_norm": k_norm_val})
                self.logger.debug("Normalized trial metrics: %s", normalized)

                # -------------------------------
                # Determine ideal (max) and anti-ideal (min) normalized values for each metric.
                # -------------------------------
                s_norm_vals = [item["s_norm"] for item in normalized]
                n_norm_vals = [item["n_norm"] for item in normalized]
                k_norm_vals = [item["k_norm"] for item in normalized]

                ideal_s = max(s_norm_vals)
                ideal_n = max(n_norm_vals)
                ideal_k = max(k_norm_vals)

                anti_s = min(s_norm_vals)
                anti_n = min(n_norm_vals)
                anti_k = min(k_norm_vals)
                self.logger.debug(
                    "Ideal values: (%.4f, %.4f, %.4f), Anti-ideal values: (%.4f, %.4f, %.4f)",
                    ideal_s, 
                    ideal_n, 
                    ideal_k, 
                    anti_s, 
                    anti_n, 
                    anti_k
                )

                # -------------------------------
                # Compute TOPSIS score for each trial by comparing distances to ideal and anti-ideal points.
                # -------------------------------
                topsised = []
                for item in normalized:
                    dist_ideal = self._euclidean_distance_3d(
                        item["s_norm"], item["n_norm"], item["k_norm"],
                        ideal_s, ideal_n, ideal_k
                    )
                    dist_anti = self._euclidean_distance_3d(
                        item["s_norm"], item["n_norm"], item["k_norm"],
                        anti_s, anti_n, anti_k
                    )
                    # Compute TOPSIS score (ratio of distance to anti-ideal over total distance).
                    topsis_score = dist_anti / (dist_ideal + dist_anti) if (dist_ideal + dist_anti) != 0 else 0
                    topsised.append({**item, "dist_ideal": dist_ideal, "dist_anti": dist_anti, "score": topsis_score})
                
                # -------------------------------
                # Log the TOPSIS scores for all Pareto trials.
                # -------------------------------
                self.logger.info("\n*** TOPSIS on Pareto front ***")
                for i, item in enumerate(sorted(topsised, key=lambda x: -x["score"]), 1):
                    self.logger.info("%d) Trial #%d - Score: %.4f", i, item['trial'].number, item['score'])
                    self.logger.info("    Silhouette: %.4f", item['silhouette'])
                    self.logger.info("    -Noise:     %.4f", item['neg_noise'])
                    self.logger.info("    -k:         %.4f", item['neg_k'])
                
                # -------------------------------
                # Select the trial with the highest TOPSIS score.
                # -------------------------------
                best_sol = max(topsised, key=lambda x: x["score"])
                best_trial = best_sol["trial"]
                self.logger.info("\nSelected by TOPSIS => Trial #%d with Score = %.4f", best_trial.number, best_sol['score'])
                self.logger.debug("Best trial selected: %s", best_trial)
                return best_trial, "pareto_topsis"
            except Exception as e:
                self.logger.error("TOPSIS failed with error: %s", str(e))
        # -------------------------------
        # If no valid Pareto solutions found, log error and raise an exception.
        # -------------------------------
        self.logger.error("No valid Pareto optimal solutions found... Raising error.")
        raise ValueError(
            "No valid solutions found during hyperparameter optimization. Try again with more data."
        )

    def optimize(self, filtered_df: pd.DataFrame):
        """
        Optimize UMAP and HDBSCAN hyperparameters and perform clustering on the input DataFrame.

        This method takes a DataFrame containing an 'embedding_vector' column defined by self.embedding_col_name, performs hyperparameter tuning using
        a triple-objective optimization (silhouette score, negative noise ratio, and negative cluster count), selects the
        best model using TOPSIS on the Pareto front, and applies the chosen models to produce clustering results.

        The final DataFrame is augmented with:
            - A single column 'reduced_vector' containing the reduced dimension coordinates as a vector for each entry.
            - Cluster labels in 'cluster_id'.
            - Membership strengths and outlier scores.

        Args:
            filtered_df (pd.DataFrame): Input DataFrame with an 'embedding_vector' column defined by self.embedding_col_name.

        Returns:
            dict: A dictionary containing:
                - 'clustered_df': The DataFrame augmented with clustering results.
                - 'umap_model': The final UMAP model instance.
                - 'hdbscan_model': The final HDBSCAN model instance.
                - 'metrics_dict': Dictionary of key metrics (reduced dimensions, number of clusters, noise ratio, and optionally silhouette score).
        """
        try:
            self.logger.debug("Starting optimization process on DataFrame with %d rows", len(filtered_df))
            # -------------------------------
            # Validate input: Check if the DataFrame contains the required embedding column.
            # -------------------------------
            if self.embedding_col_name not in filtered_df.columns:
                self.logger.error(f"Input DataFrame must contain a(n) {self.embedding_col_name} column.")
                raise ValueError(f"Missing {self.embedding_col_name} column in input DataFrame.")

            # -------------------------------
            # Convert embedding vectors to a 2D numpy array.
            # -------------------------------
            embeddings = np.vstack(filtered_df[self.embedding_col_name].values) #type: ignore
            num_data_pts = len(filtered_df)
            self.logger.debug("Converted embedding vectors to numpy array with shape: %s", embeddings.shape)

            self.logger.info("Starting triple-objective optimization (silhouette, -noise, -k).")
            # -------------------------------
            # Create an Optuna study configured for multi-objective optimization.
            # -------------------------------
            study = optuna.create_study(
                directions=["maximize", "maximize", "maximize"],
                sampler=optuna.samplers.NSGAIISampler(seed=self.random_state),
            )
            self.logger.debug("Optuna study created with directions maximize for all objectives.")
            
            # -------------------------------
            # Run the hyperparameter optimization for a specified number of trials.
            # -------------------------------
            study.optimize(
                lambda trial: self._triple_objective(trial, embeddings),
                n_trials=self.n_trials,
                n_jobs=self.optuna_jobs,
                show_progress_bar=True,
            )
            self.logger.debug("Optuna optimization completed.")

            # -------------------------------
            # Extract Pareto-optimal trials with valid metric values.
            # -------------------------------
            pareto_trials = [t for t in study.best_trials if not any(math.isinf(x) for x in t.values)]
            self.logger.info("\nPareto front trials:")
            self.logger.info("Number of Pareto-optimal solutions: %d", len(pareto_trials))
            # -------------------------------
            # Log details for each Pareto trial.
            # -------------------------------
            for i, trial in enumerate(pareto_trials, 1):
                s_val, neg_noise_val, neg_k_val = trial.values
                self.logger.info("\nSolution %d:", i)
                self.logger.info("    - clusters: %d", int(-neg_k_val))
                self.logger.info("    - silhouette: %.3f", s_val)
                self.logger.info("    - noise ratio: %.3f", -neg_noise_val)
                self.logger.debug("Trial #%d details: values=%s, params=%s", trial.number, trial.values, trial.params)

            # -------------------------------
            # Select the best trial using TOPSIS or fallback to default hyperparameters if no valid trial is found.
            # -------------------------------
            if not pareto_trials:
                self.logger.warning("No valid Pareto-optimal solutions found; falling back to average of min/max hyperparameters.")
                best_umap, best_hdbscan, umap_params, hdbscan_params = self._default_models(num_data_pts)
                dims_final = umap_params["n_components"]
            else:
                # -------------------------------
                # Use TOPSIS to select the best trial from the Pareto front.
                # -------------------------------
                best_trial, method_used = self._get_best_solution(study, pareto_trials)
                self.logger.info("Solution selection method: %s", method_used)
                s_val, neg_noise_val, neg_k_val = best_trial.values
                self.logger.info("\n*** Final Chosen Trial ***")
                self.logger.info(" - Silhouette: %.4f", s_val)
                self.logger.info(" - Neg noise:  %.4f", neg_noise_val)
                self.logger.info(" - Neg k:      %.4f", neg_k_val)
                self.logger.debug("Best trial parameters: %s", best_trial.params)
                dims_final = self.dims if self.dims is not None else best_trial.params["umap_n_components"]

                best_umap, best_hdbscan, umap_params, hdbscan_params = self._create_models(best_trial, num_data_pts)
            
            self.logger.debug("Using final UMAP dims=%d", dims_final)
            # -------------------------------
            # Fit UMAP on embeddings to obtain reduced dimension coordinates.
            # -------------------------------
            reduced_coords = best_umap.fit_transform(embeddings)
            self.logger.debug("UMAP reduction complete. Reduced coordinates shape: %s", reduced_coords.shape) #type: ignore
            
            # -------------------------------
            # Predict clusters using the HDBSCAN model.
            # -------------------------------
            cluster_id = best_hdbscan.fit_predict(reduced_coords) #type: ignore
            self.logger.debug("HDBSCAN produced cluster numbers: %s", cluster_id)

            # -------------------------------
            # Validate that the length of membership probabilities matches the DataFrame.
            # -------------------------------
            if len(best_hdbscan.probabilities_) != len(filtered_df):
                raise AssertionError(
                    f"Mismatch between probabilities ({len(best_hdbscan.probabilities_)}) and dataframe length ({len(filtered_df)})"
                )
            # -------------------------------
            # Append clustering metadata (membership strengths, core point, outlier scores) to the DataFrame.
            # -------------------------------
            filtered_df["membership_strength"] = best_hdbscan.probabilities_
            outlier_scores = best_hdbscan.outlier_scores_
            threshold = np.percentile(outlier_scores, self.hdbscan_outlier_threshold)
            filtered_df["core_point"] = outlier_scores < threshold
            filtered_df["outlier_score"] = outlier_scores
            self.logger.debug("Appended membership_strength, core_point, and outlier_score to DataFrame.")

            # -------------------------------
            # Compute overall clustering metrics (noise ratio, number of clusters, silhouette score).
            # -------------------------------
            noise_ratio = (cluster_id == -1).sum() / len(cluster_id)
            n_clusters = len(set(cluster_id)) - (1 if -1 in cluster_id else 0)

            best_sil_score = None
            mask = cluster_id != -1
            if np.sum(mask) >= 2:
                best_sil_score = silhouette_score(
                    reduced_coords[mask], #type: ignore
                    cluster_id[mask],
                    metric="euclidean",
                )
                self.logger.debug("Computed silhouette score for non-noise points: %.4f", best_sil_score)

            # -------------------------------
            # Log final clustering results and cluster sizes.
            # -------------------------------
            self.logger.info("\n*** Final Clustering Results ***:")
            self.logger.info("Dimensionality: %d", dims_final)
            self.logger.info("Number of clusters: %d", n_clusters)
            if best_sil_score is not None:
                self.logger.info("Silhouette score: %.3f", best_sil_score)
            self.logger.info("Noise ratio: %.1f%%", noise_ratio * 100)

            unique_labels = sorted(set(cluster_id))
            if -1 in unique_labels:
                unique_labels.remove(-1)
            self.logger.info("\nCluster sizes:")
            for label in unique_labels:
                size = (cluster_id == label).sum()
                self.logger.info("  Cluster %d: %d points (%.1f%%)", label, size, size / len(cluster_id) * 100)
                self.logger.debug("Cluster %d size: %d", label, size)

            # -------------------------------
            # Add the reduced dimension vector as a single column in the DataFrame.
            # -------------------------------
            filtered_df["reduced_vector"] = [list(row) for row in reduced_coords]
            self.logger.debug("Added column 'reduced_vector' with reduced dimension vectors.")
            filtered_df["cluster_id"] = cluster_id

            # -------------------------------
            # Build a dictionary containing key clustering metrics.
            # -------------------------------
            metrics_dict = {
                "reduced_dimensions": dims_final,
                "n_clusters": n_clusters,
                "noise_ratio": round(float(noise_ratio), 2),
            }
            if best_sil_score is not None:
                metrics_dict["silhouette_score"] = round(float(best_sil_score), 2)
            self.logger.debug("Built metrics_dict: %s", metrics_dict)

            # -------------------------------
            # Finalize and return the clustering results and models.
            # -------------------------------
            self.logger.debug("Optimization process complete. Returning final results.")
            return {
                "clustered_df": filtered_df,
                "umap_model": best_umap,
                "hdbscan_model": best_hdbscan,
                "metrics_dict": metrics_dict,
            }
        except Exception as e:
            self.logger.error("An error occurred during clustering optimization: %s", str(e))
            raise


def run_clustering(
    filtered_df: pd.DataFrame,
    dims=3, 
    min_clusters=3,
    max_clusters=25,
    n_trials=50, 
    random_state=None,
    embedding_col_name="embedding_vector",
    
    umap_n_neighbors_min=2,
    umap_n_neighbors_max=25,
    umap_min_dist_min=0.0,
    umap_min_dist_max=0.1,
    umap_spread_min=1.0,
    umap_spread_max=10.0,
    umap_learning_rate_min=0.08,
    umap_learning_rate_max=1.0,
    umap_min_dims=2,
    umap_max_dims=20,
    umap_metric="cosine",
    
    hdbscan_min_cluster_size_multiplier_min=0.005,
    hdbscan_min_cluster_size_multiplier_max=0.025,
    hdbscan_min_samples_min=2,
    hdbscan_min_samples_max=50,
    hdbscan_epsilon_min=0.0,
    hdbscan_epsilon_max=1.0,
    hdbscan_metric="euclidean",
    hdbscan_cluster_selection_method="eom",
    
    optuna_jobs=-1,
    hdbscan_outlier_threshold=90,
    log_level=logging.INFO
):
    """
    Perform clustering on a DataFrame containing embedding vectors.

    This function is a convenient functional interface for using the ClusteringEngine.
    It expects the DataFrame to have an 'embedding_vector' column defined by self.embedding_col_name, and it returns a dictionary with:
        - 'clustered_df': The DataFrame augmented with clustering results (including reduced dimensions, cluster labels, etc.).
        - 'umap_model': The UMAP model instance used for dimensionality reduction.
        - 'hdbscan_model': The HDBSCAN model instance used for clustering.
        - 'metrics_dict': A dictionary containing key clustering metrics.

    Args:
        filtered_df (pd.DataFrame): DataFrame containing the embedding vectors in 'embedding_vector' column defined by self.embedding_col_name.
        log_level: Logging level (e.g., logging.INFO, logging.DEBUG)

    Returns:
        dict: Dictionary containing clustering results, models, and metrics.
        
    The clustered_df that is returned has the columns: 
    'content', 'embedding_vector', 'membership_strength', 'core_point', 'outlier_score', 'reduced_vector', 'cluster_id'
    """
    try:
        logging.debug("cluster() function called.")
        # -------------------------------
        # Create an instance of ClusteringEngine and start the optimization process.
        # -------------------------------
        clustering = ClusteringEngine(
            dims=dims,
            min_clusters=min_clusters,
            max_clusters=max_clusters,
            n_trials=n_trials,
            random_state=random_state,
            embedding_col_name=embedding_col_name,
            umap_n_neighbors_min=umap_n_neighbors_min,
            umap_n_neighbors_max=umap_n_neighbors_max,
            umap_min_dist_min=umap_min_dist_min,
            umap_min_dist_max=umap_min_dist_max,
            umap_spread_min=umap_spread_min,
            umap_spread_max=umap_spread_max,
            umap_learning_rate_min=umap_learning_rate_min,
            umap_learning_rate_max=umap_learning_rate_max,
            umap_min_dims=umap_min_dims,
            umap_max_dims=umap_max_dims,
            umap_metric=umap_metric,
            hdbscan_min_cluster_size_multiplier_min=hdbscan_min_cluster_size_multiplier_min,
            hdbscan_min_cluster_size_multiplier_max=hdbscan_min_cluster_size_multiplier_max,
            hdbscan_min_samples_min=hdbscan_min_samples_min,
            hdbscan_min_samples_max=hdbscan_min_samples_max,
            hdbscan_epsilon_min=hdbscan_epsilon_min,
            hdbscan_epsilon_max=hdbscan_epsilon_max,
            hdbscan_metric=hdbscan_metric,
            hdbscan_cluster_selection_method=hdbscan_cluster_selection_method,
            optuna_jobs=optuna_jobs,
            hdbscan_outlier_threshold=hdbscan_outlier_threshold
        )
        logging.debug("ClusteringEngine instance created. Calling optimize().")
        result = clustering.optimize(filtered_df=filtered_df)
        logging.debug("Clustering process completed successfully.")
        return result
    except Exception as e:
        logging.error("An error occurred in the clustering process: %s", str(e))
        raise