from atmose.clustering import run_clustering
from atmose.labeling import add_labels
from atmose.plotting import silhouette_fig, scatter_fig, bars_fig
import plotly.io as pio
import pandas as pd
import logging

class Pipeline:
    def __init__(self, 
        data, 
        environment=None,
        dims=3, 
        min_clusters=3, 
        max_clusters=25, 
        n_trials=50, 
        random_state=None,
        embedding_col_name="embedding_vector",
        
        umap_n_neighbors_min=2,
        umap_n_neighbors_max=25,
        umap_min_dist_min=0.0,
        umap_min_dist_max=0.1,
        umap_spread_min=1.0,
        umap_spread_max=10.0,
        umap_learning_rate_min=0.08,
        umap_learning_rate_max=1.0,
        umap_min_dims=2,
        umap_max_dims=20,
        umap_metric="cosine",
        
        hdbscan_min_cluster_size_multiplier_min=0.005,
        hdbscan_min_cluster_size_multiplier_max=0.025,
        hdbscan_min_samples_min=2,
        hdbscan_min_samples_max=50,
        hdbscan_epsilon_min=0.0,
        hdbscan_epsilon_max=1.0,
        hdbscan_metric="euclidean",
        hdbscan_cluster_selection_method="eom",
        
        optuna_jobs=-1,
        hdbscan_outlier_threshold=90,
        log_level=logging.INFO,
        
        llm_model='o1', 
        language='english', 
        ascending=False, 
        core_top_n=12, 
        peripheral_n=12,
        llm_temp=1.0,
        num_strata=3,
        content_col_name='content',
        data_description="No data description provided. Just do your best to infer/assume the context of the data while performing your tasks.",
        
        wrap_width: int = 100,
        
    ):
        self.data = data
        # pio.renderers.default = environment
        self.environment = environment
        print(f"Using {environment} environment for rendering plots. Use None for default, 'colab' for Google Colab, or 'notebook' for Jupyter Notebook.")
        self.dims = dims
        self.min_clusters = min_clusters
        self.max_clusters = max_clusters
        self.n_trials = n_trials
        self.random_state = random_state
        self.embedding_col_name = embedding_col_name
        self.umap_n_neighbors_min = umap_n_neighbors_min
        self.umap_n_neighbors_max = umap_n_neighbors_max
        self.umap_min_dist_min = umap_min_dist_min
        self.umap_min_dist_max = umap_min_dist_max
        self.umap_spread_min = umap_spread_min
        self.umap_spread_max = umap_spread_max
        self.umap_learning_rate_min = umap_learning_rate_min
        self.umap_learning_rate_max = umap_learning_rate_max
        self.umap_min_dims = umap_min_dims
        self.umap_max_dims = umap_max_dims
        self.umap_metric = umap_metric
        self.hdbscan_min_cluster_size_multiplier_min = hdbscan_min_cluster_size_multiplier_min
        self.hdbscan_min_cluster_size_multiplier_max = hdbscan_min_cluster_size_multiplier_max
        self.hdbscan_min_samples_min = hdbscan_min_samples_min
        self.hdbscan_min_samples_max = hdbscan_min_samples_max
        self.hdbscan_epsilon_min = hdbscan_epsilon_min
        self.hdbscan_epsilon_max = hdbscan_epsilon_max
        self.hdbscan_metric = hdbscan_metric
        self.hdbscan_cluster_selection_method = hdbscan_cluster_selection_method
        self.optuna_jobs = optuna_jobs
        self.hdbscan_outlier_threshold = hdbscan_outlier_threshold
        self.log_level = log_level
        
        self.llm_model = llm_model
        self.language = language
        self.ascending = ascending
        self.core_top_n = core_top_n
        self.peripheral_n = peripheral_n
        self.llm_temp = llm_temp
        self.num_strata = num_strata
        self.content_col_name = content_col_name
        self.data_description = data_description
        
        self.wrap_width = wrap_width
        

    def _run(self):
        clustered_df = run_clustering(
            filtered_df=self.data,
            dims=self.dims,
            min_clusters=self.min_clusters,
            max_clusters=self.max_clusters,
            n_trials=self.n_trials,
            random_state=self.random_state,
            embedding_col_name=self.embedding_col_name,
            umap_n_neighbors_min=self.umap_n_neighbors_min,
            umap_n_neighbors_max=self.umap_n_neighbors_max,
            umap_min_dist_min=self.umap_min_dist_min,
            umap_min_dist_max=self.umap_min_dist_max,
            umap_spread_min=self.umap_spread_min,
            umap_spread_max=self.umap_spread_max,
            umap_learning_rate_min=self.umap_learning_rate_min,
            umap_learning_rate_max=self.umap_learning_rate_max,
            umap_min_dims=self.umap_min_dims,
            umap_max_dims=self.umap_max_dims,
            umap_metric=self.umap_metric,
            hdbscan_min_cluster_size_multiplier_min=self.hdbscan_min_cluster_size_multiplier_min,
            hdbscan_min_cluster_size_multiplier_max=self.hdbscan_min_cluster_size_multiplier_max,
            hdbscan_min_samples_min=self.hdbscan_min_samples_min,
            hdbscan_min_samples_max=self.hdbscan_min_samples_max,
            hdbscan_epsilon_min=self.hdbscan_epsilon_min,
            hdbscan_epsilon_max=self.hdbscan_epsilon_max,
            hdbscan_metric=self.hdbscan_metric,
            hdbscan_cluster_selection_method=self.hdbscan_cluster_selection_method,
            optuna_jobs=self.optuna_jobs,
            hdbscan_outlier_threshold=self.hdbscan_outlier_threshold,
            log_level=self.log_level
        )['clustered_df']
        
        labeled_df = add_labels(
            cluster_df=clustered_df,
            llm_model=self.llm_model,
            language=self.language,
            ascending=self.ascending,
            core_top_n=self.core_top_n,
            peripheral_n=self.peripheral_n,
            llm_temp=self.llm_temp,
            num_strata=self.num_strata,
            content_col_name=self.content_col_name,
            data_description=self.data_description,
            log_level=self.log_level
        )
        
        # Create figures
        sil_fig = silhouette_fig(
            clustered_df=labeled_df,
            content_col_name=self.content_col_name
        )
        sca_fig = scatter_fig(
            clusters_df=labeled_df,
            wrap_width=self.wrap_width,
            content_col_name=self.content_col_name
        )
        bar_fig = bars_fig(
            clusters_df=labeled_df,
            content_col_name=self.content_col_name
        )
        
        sil_fig.show(renderer=f'"{self.environment}"')
        sca_fig.show(renderer=f'"{self.environment}"')
        bar_fig.show(renderer=f'"{self.environment}"')
        print(f'using renderer="{self.environment}"')

def full_pipeline(
    data: pd.DataFrame,
    environment: str|None = None,
    dims=3, 
    min_clusters=3, 
    max_clusters=25, 
    n_trials=50, 
    random_state=None,
    embedding_col_name="embedding_vector",
    
    umap_n_neighbors_min=2,
    umap_n_neighbors_max=25,
    umap_min_dist_min=0.0,
    umap_min_dist_max=0.1,
    umap_spread_min=1.0,
    umap_spread_max=10.0,
    umap_learning_rate_min=0.08,
    umap_learning_rate_max=1.0,
    umap_min_dims=2,
    umap_max_dims=20,
    umap_metric="cosine",
    
    hdbscan_min_cluster_size_multiplier_min=0.005,
    hdbscan_min_cluster_size_multiplier_max=0.025,
    hdbscan_min_samples_min=2,
    hdbscan_min_samples_max=50,
    hdbscan_epsilon_min=0.0,
    hdbscan_epsilon_max=1.0,
    hdbscan_metric="euclidean",
    hdbscan_cluster_selection_method="eom",
    
    optuna_jobs=-1,
    hdbscan_outlier_threshold=90,
    log_level=logging.INFO,
    
    llm_model='o1', 
    language='english', 
    ascending=False, 
    core_top_n=12, 
    peripheral_n=12,
    llm_temp=1.0,
    num_strata=3,
    content_col_name='content',
    data_description="No data description provided. Just do your best to infer/assume the context of the data while performing your tasks.",
    
    wrap_width: int = 100
):
    """
    Performs the full EDA pipeline on the given data- dimensionality reduction, clustering, labeling, and plotting.
    """
    pipeline = Pipeline(
        data=data,
        environment=environment,
        dims=dims,
        min_clusters=min_clusters,
        max_clusters=max_clusters,
        n_trials=n_trials,
        random_state=random_state,
        embedding_col_name=embedding_col_name,
        umap_n_neighbors_min=umap_n_neighbors_min,
        umap_n_neighbors_max=umap_n_neighbors_max,
        umap_min_dist_min=umap_min_dist_min,
        umap_min_dist_max=umap_min_dist_max,
        umap_spread_min=umap_spread_min,
        umap_spread_max=umap_spread_max,
        umap_learning_rate_min=umap_learning_rate_min,
        umap_learning_rate_max=umap_learning_rate_max,
        umap_min_dims=umap_min_dims,
        umap_max_dims=umap_max_dims,
        umap_metric=umap_metric,
        hdbscan_min_cluster_size_multiplier_min=hdbscan_min_cluster_size_multiplier_min,
        hdbscan_min_cluster_size_multiplier_max=hdbscan_min_cluster_size_multiplier_max,
        hdbscan_min_samples_min=hdbscan_min_samples_min,
        hdbscan_min_samples_max=hdbscan_min_samples_max,
        hdbscan_epsilon_min=hdbscan_epsilon_min,
        hdbscan_epsilon_max=hdbscan_epsilon_max,
        hdbscan_metric=hdbscan_metric,
        hdbscan_cluster_selection_method=hdbscan_cluster_selection_method,
        optuna_jobs=optuna_jobs,
        hdbscan_outlier_threshold=hdbscan_outlier_threshold,
        log_level=log_level,
        llm_model=llm_model,
        language=language,
        ascending=ascending,
        core_top_n=core_top_n,
        peripheral_n=peripheral_n,
        llm_temp=llm_temp,
        num_strata=num_strata,
        content_col_name=content_col_name,
        data_description=data_description,
        wrap_width=wrap_width
    )
    pipeline._run()